/*
 * Created by Brett on 17/11/23.
 * Licensed under GNU General Public License V3.0
 * See LICENSE file for license detail
 */
#include <bf_interpreter.h>
#include <iostream>
#include <functional>
#include <utility>

template<typename functor>
static inline void match(functor f, int sp, size_t& index, const std::string& program)
{
    while (f(index) < program.size())
    {
        if (program[index] == '[')
            sp++;
        if (program[index] == ']')
            sp--;
        if (sp == 0)
            break;
    }
}

template<typename functor>
static inline void match(functor f, int sp, size_t& index, const std::vector<bf_token_t>& program)
{
    while (f(index) < program.size())
    {
        if (program[index].type == bf_token::OPEN)
            sp++;
        if (program[index].type == bf_token::CLOSE)
            sp--;
        if (sp == 0)
            break;
    }
}

void interpret_bf(const std::string& program)
{
    brainfuck_interpreter fuck;
    
    size_t index = 0;
    while (index < program.size())
    {
        auto c = program[index];
        switch (c)
        {
            case '>':
                fuck.increment_dp();
                break;
            case '<':
                fuck.decrement_dp();
                break;
            case '+':
                fuck.increment();
                break;
            case '-':
                fuck.decrement();
                break;
            case '.':
                fuck.print(std::cout);
                break;
            case ',':
                fuck.read(std::cin);
                break;
            case '[':
                if (fuck.is() == 0)
                    match([](size_t& idx) { return ++idx; }, 1, index, program);
                break;
            case ']':
                if (fuck.is() != 0)
                    match([](size_t& idx) { return --idx; }, -1, index, program);
                break;
            default:
                break;
        }
        index++;
    }
}

std::function<void(brainfuck_interpreter& inter, size_t offset, size_t& index, const std::vector<bf_token_t>& tokens)> funcs[8] = {
        [](brainfuck_interpreter& inter, size_t off, size_t& index, const std::vector<bf_token_t>& tokens) { inter.increment_dp(off); },
        [](brainfuck_interpreter& inter, size_t off, size_t& index, const std::vector<bf_token_t>& tokens) { inter.decrement_dp(off); },
        [](brainfuck_interpreter& inter, size_t off, size_t& index, const std::vector<bf_token_t>& tokens) {
            inter.increment(static_cast<int8_t>(off));
        },
        [](brainfuck_interpreter& inter, size_t off, size_t& index, const std::vector<bf_token_t>& tokens) {
            inter.decrement(static_cast<int8_t>(off));
        },
        [](brainfuck_interpreter& inter, size_t off, size_t& index, const std::vector<bf_token_t>& tokens) { inter.print(std::cout); },
        [](brainfuck_interpreter& inter, size_t off, size_t& index, const std::vector<bf_token_t>& tokens) { inter.read(std::cin); },
        [](brainfuck_interpreter& inter, size_t off, size_t& index, const std::vector<bf_token_t>& tokens) {
            if (inter.is() == 0)
                match([](size_t& idx) { return ++idx; }, 1, index, tokens);
        },
        [](brainfuck_interpreter& inter, size_t off, size_t& index, const std::vector<bf_token_t>& tokens) {
            if (inter.is() != 0)
                match([](size_t& idx) { return --idx; }, -1, index, tokens);
        },
};

void interpret_bf(bf_tokenizer& tokenizer)
{
    brainfuck_interpreter fuck;
    auto& tokens = tokenizer.data();
    
    size_t index = 0;
    while (index < tokens.size())
    {
        auto c = tokens[index];
        switch (c.type)
        {
            case bf_token::INC_DP:
                fuck.increment_dp(c.offset);
                break;
            case bf_token::DEC_DP:
                fuck.decrement_dp(c.offset);
                break;
            case bf_token::INC_DV:
                fuck.increment(static_cast<int8_t>(c.offset));
                break;
            case bf_token::DEC_DV:
                fuck.decrement(static_cast<int8_t>(c.offset));
                break;
            case bf_token::PRINT:
                fuck.print(std::cout);
                break;
            case bf_token::READ:
                fuck.read(std::cin);
                break;
            case bf_token::OPEN:
                if (fuck.is() == 0)
                    index = c.offset;
                break;
            case bf_token::CLOSE:
                if (fuck.is() != 0)
                    index = c.offset;
                break;
        }
        index++;
    }
}

void interpret_bf_dynamic(bf_tokenizer& tokenizer)
{
    brainfuck_interpreter fuck;
    auto& tokens = tokenizer.data();
    
    size_t index = 0;
    while (index < tokens.size())
    {
        auto& c = tokens[index];
        funcs[static_cast<std::underlying_type_t<bf_token>>(c.type)](fuck, c.offset, index, tokens);
        index++;
    }
}

void interpret_bf_matching(bf_tokenizer& tokenizer)
{
    brainfuck_interpreter fuck;
    auto& tokens = tokenizer.data();
    
    size_t index = 0;
    while (index < tokens.size())
    {
        auto c = tokens[index];
        switch (c.type)
        {
            case bf_token::INC_DP:
                fuck.increment_dp(c.offset);
                break;
            case bf_token::DEC_DP:
                fuck.decrement_dp(c.offset);
                break;
            case bf_token::INC_DV:
                fuck.increment(static_cast<int8_t>(c.offset));
                break;
            case bf_token::DEC_DV:
                fuck.decrement(static_cast<int8_t>(c.offset));
                break;
            case bf_token::PRINT:
                fuck.print(std::cout);
                break;
            case bf_token::READ:
                fuck.read(std::cin);
                break;
            case bf_token::OPEN:
                if (fuck.is() == 0)
                    match([](size_t& idx) { return ++idx; }, 1, index, tokens);
                break;
            case bf_token::CLOSE:
                if (fuck.is() != 0)
                    match([](size_t& idx) { return --idx; }, -1, index, tokens);
                break;
        }
        index++;
    }
}

void brainfuck_interpreter::check_size()
{
    if (_dp >= _size)
    {
        char* tmp = new char[_size * 2];
        std::memset(&tmp[_size], 0, _size);
        std::memcpy(tmp, _data, _size);
        _size = _size * 2;
        delete[] _data;
        _data = tmp;
    }
}
